# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-09-02 20:29
from __future__ import unicode_literals
import os

from graphviz import Digraph

from django.db import migrations
from django.core.files.base import ContentFile
from django.contrib.contenttypes.management import create_contenttypes

from bayesian_networks.bayespy_constants import (
    DIST_GAUSSIAN_ARD, DIST_GAMMA)
from bayesian_networks.models import BayesianNetworkNode as BNN


def generate_bn_image(bn):
    """
    Auxiliary function for generating the image of a BN, as model
    methods are not available in migrations.
    """
    dot = Digraph(comment=bn.name)
    nodes = bn.nodes.all()
    for node in nodes:
        dot.node(name=node.name, label=node.name)
    edges = bn.edges.all()
    for edge in edges:
         dot.edge(str(edge.parent.name),
                 str(edge.child.name))
    dot.format = "png"
    contentfile = ContentFile(dot.pipe())
    image_name = "{0}/{1}".format(
            os.path.join("django_ai",
                         "bayesian_networks"),
            bn.name + ".png")
    bn.image.save(image_name, contentfile)
    bn.save()


def create_bn1_example(apps, schema_editor):
    # Content Types Hackery for ensuring that it exists
    app_config = apps.get_app_config('examples')
    app_config.models_module = app_config.models_module or True
    create_contenttypes(app_config)
    ##

    BayesianNetwork = apps.get_model("bayesian_networks",
                                     "BayesianNetwork")
    BayesianNetworkEdge = apps.get_model("bayesian_networks",
                                         "BayesianNetworkEdge")
    BayesianNetworkNode = apps.get_model("bayesian_networks",
                                         "BayesianNetworkNode")
    ContentType = apps.get_model("contenttypes",
                                "ContentType")
    
    bn1 = BayesianNetwork(name="BN1 (Example)")
    bn1.save()
    mu = BayesianNetworkNode(
        network=bn1,
        name="mu",
        node_type=BNN.NODE_TYPE_STOCHASTIC,
        is_observable=False,
        distribution=DIST_GAUSSIAN_ARD,
        distribution_params="0, 1e-6",
        graph_interval="-10, 20"
    )
    tau = BayesianNetworkNode(
        network=bn1,
        name="tau",
        node_type=BNN.NODE_TYPE_STOCHASTIC,
        is_observable=False,
        distribution=DIST_GAMMA,
        distribution_params="1e-6, 1e-6",
        graph_interval="1e-6, 0.1"
    )
    ui_avg1 = BayesianNetworkNode(
        network=bn1,
        name="userinfo.avg1",
        node_type=BNN.NODE_TYPE_STOCHASTIC,
        is_observable=True,
        distribution=DIST_GAUSSIAN_ARD,
        distribution_params="mu, tau",
        ref_model=ContentType.objects.get(model="userinfo",
                                          app_label="examples"),
        ref_column="avg1",

    )
    mu.save()
    tau.save()
    ui_avg1.save()

    mu_to_ui_avg1 = BayesianNetworkEdge(
        network=bn1,
        description="mu -> userinfo.avg1",
        parent=mu,
        child=ui_avg1
    )
    tau_to_ui_avg1 = BayesianNetworkEdge(
        network=bn1,
        description="tau -> userinfo.avg1",
        parent=tau,
        child=ui_avg1
    )
    mu_to_ui_avg1.save()
    tau_to_ui_avg1.save()
    # Generate the image
    generate_bn_image(bn1)


def delete_bn1_example(apps, schema_editor):
    BayesianNetwork = apps.get_model("bayesian_networks",
                                     "BayesianNetwork")

    BayesianNetwork.objects.get(name="BN1 (Example)").delete()


class Migration(migrations.Migration):

    dependencies = [
        ('examples', '0003_populate_userinfo'),
    ]

    operations = [
        migrations.RunPython(create_bn1_example,
                             delete_bn1_example),
    ]


# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-05-02 23:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('examples', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='userinfo',
            name='avg1',
            field=models.FloatField(blank=True, null=True, verbose_name='Average 1'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-05-02 23:12
from __future__ import unicode_literals
import numpy as np

from django.db import migrations


def populate_userinfos(apps, schema_editor):
    UserInfo = apps.get_model("examples", "UserInfo")
    # Use a fixed seed for generate content
    np.random.seed(123456)
    # Size of table
    size = 200
    # Sex is ~ 70% F (0) / 30% M (1)
    sex = np.random.binomial(1, 0.7, size)  # 200 Bernoullies :)
    # Age is around 30, mostly between 25 and 35
    age = np.floor(np.random.normal(30, 2, size=(size,)))
    # Average 1 is a metric normally distributed around 10 with a std dev of 5
    avg1 = np.random.normal(10, 5, size=(size,))
    # Create the objects in the Model
    uis = []
    for i in range(0, size):
        uis.append(UserInfo(age=age[i], sex=sex[i], avg1=avg1[i]))
    UserInfo.objects.bulk_create(uis)


def unpopuplate_userinfos(apps, schema_editor):
    UserInfo = apps.get_model("examples", "UserInfo")
    UserInfo.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('examples', '0002_student_avg1'),
    ]

    operations = [
        migrations.RunPython(populate_userinfos,
                             unpopuplate_userinfos),
    ]



# -*- coding: utf-8 -*-

import os

import matplotlib
matplotlib.use('Agg')

from graphviz import Digraph
import numpy as np
import bayespy as bp

from django.db import models
from django.core.files.base import ContentFile
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone
from django.db.models.signals import post_save
from django.conf import settings

from django_dag.models import node_factory, edge_factory
from picklefield.fields import PickledObjectField

from .bayespy_constants import (DISTRIBUTION_CHOICES,
                                DETERMINISTIC_CHOICES)


class BayesianNetwork(models.Model):
    """
    Model for a Bayesian Network
    """
    _Q = None

    name = models.CharField("Name", max_length=100)
    image = models.ImageField("Image", blank=True, null=True)
    engine_object = PickledObjectField(blank=True, null=True)
    engine_object_timestamp = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return("BN: " + self.name)

    def save(self, *args, **kwargs):
        """
        """
        super(BayesianNetwork, self).save(*args, **kwargs)

    def get_graph(self):
        dot = Digraph(comment=self.name)
        nodes = self.nodes.all()
        for node in nodes:
            dot.node(name=node.name, label=node.name)
        edges = self.edges.all()
        for edge in edges:
            # Pending: debug constraint param
            dot.edge(str(edge.parent.name),
                     str(edge.child.name))
        return(dot)

    def get_raw_network_image(self):
        dot = self.get_graph()
        dot.format = "png"
        return ContentFile(dot.pipe())

    def update_image(self):
        if self.image:
            self.image.delete()
        image_name = "{0}/{1}".format(
            os.path.join("django_ai",
                         "bayesian_networks"),
            self.name + ".png")
        self.image.save(image_name,
                        self.get_raw_network_image())

    def get_nodes_names(self):
        return(self.nodes.all().values_list("name", flat=True))

    def get_engine_object(self, reconstruct=False, save=True):
        if self.engine_object and not reconstruct:
            return(self.engine_object)

        nodes_eos = []
        nodes = self.nodes.all()
        nodes_dict = {n.name: n for n in nodes}
        root_nodes = {n.name: n.get_engine_object(reconstruct=reconstruct)
                      for n in nodes if not n.parents()}
        child_nodes = [n for n in nodes if n.parents()]
        # TODO: Currently support for one level of depth, pending for any graph
        for node_name in root_nodes:
            nodes_eos.append(root_nodes[node_name])
        for c_n in child_nodes:
            p_ns = {n: root_nodes[n] for n in c_n.parse_nodes_in_dist_params()}
            nodes_eos.append(c_n.get_engine_object(
                parents=p_ns, reconstruct=reconstruct))
        self.engine_object = bp.inference.VB(*nodes_eos)

        if save:
            self.engine_object_timestamp = timezone.now()
            self.save()
        return(self.engine_object)

    def perform_inference(self, iters=100, recalculate=False, save=True):
        if not self.engine_object or recalculate:
            Q = self.get_engine_object(reconstruct=True)
            # Run the inference
            Q.update(repeat=iters)
            if save:
                self.engine_object_timestamp = timezone.now()
                self.save()
                # Propagate results to the network
                for node in self.nodes.all():
                    node.engine_inferred_object = Q[node.name]
                    node.engine_object_timestamp = timezone.now()
                    node.update_image()
                    node.save()
        return(self.engine_object)

    def get_inf_object(self):
        return(self.engine_object)

    def reset_engine_object(self, save=True):
        self.engine_object = None
        self.engine_object_timestamp = None
        if save:
            self.save()
        return(True)

    def reset_inference(self, save=True):
        self.reset_engine_object(save=save)
        for node in self.nodes.all():
            node.reset_inference(save=save)
        return(True)

    @property
    def is_inferred(self):
        return(self.engine_object is not None)


class BayesianNetworkEdge(
        edge_factory('bayesian_networks.BayesianNetworkNode')):
    """
    An Edge between Bayesian Network Nodes
    """
    network = models.ForeignKey(BayesianNetwork, related_name="edges")
    description = models.CharField("Description", max_length=50)

    def __str__(self):
        return(self.description)

    def save(self, *args, **kwargs):
        self.network = self.parent.network
        super(BayesianNetworkEdge, self).save(*args, **kwargs)


class BayesianNetworkNode(
        node_factory('bayesian_networks.BayesianNetworkEdge')):
    """
    A Node in a Bayesian Network
    """
    _engine_object = None  # Deprecated?
    _engine_inferred_object = None

    NODE_TYPE_STOCHASTIC = 0
    NODE_TYPE_DETERMINISTIC = 1
    NODE_TYPE_CHOICES = (
        (NODE_TYPE_STOCHASTIC, "Stochastic"),
        (NODE_TYPE_DETERMINISTIC, "Deterministic"),
    )

    network = models.ForeignKey(BayesianNetwork, related_name="nodes")
    name = models.CharField("Node Name", max_length=50)
    node_type = models.SmallIntegerField("Type", choices=NODE_TYPE_CHOICES,
                                         default=NODE_TYPE_STOCHASTIC)
    is_observable = models.BooleanField("Is Observable?", default=True)
    distribution = models.CharField("Distribution", max_length=50,
                                    choices=DISTRIBUTION_CHOICES,
                                    blank=True, null=True)
    distribution_params = models.CharField("Distribution Parameters",
                                           max_length=200,
                                           blank=True, null=True)
    deterministic = models.CharField("Deterministic", max_length=50,
                                     choices=DETERMINISTIC_CHOICES,
                                     blank=True, null=True)
    deterministic_params = models.CharField("Deterministic Parameters",
                                            max_length=200,
                                            blank=True, null=True)
    ref_model = models.ForeignKey(ContentType, on_delete=models.CASCADE,
                                  blank=True, null=True)
    ref_column = models.CharField("Reference Column", max_length=100,
                                  blank=True, null=True)
    engine_object = PickledObjectField(blank=True, null=True)
    engine_object_timestamp = models.DateTimeField(blank=True, null=True)
    engine_inferred_object = PickledObjectField(blank=True, null=True)
    engine_inferred_object_timestamp = models.DateTimeField(
        blank=True, null=True)
    graph_interval = models.CharField("Graph Interval", max_length=20,
                                      blank=True, null=True)
    image = models.ImageField("Image", blank=True, null=True)

    def get_data(self):
        if not self.is_observable:
            return(False)
        data = self.ref_model.model_class().objects.values_list(
            self.ref_column, flat=True)
        return(data)

    def get_parents_names(self):
        return(self.parents.objects.values_list("name", flat=True))

    def parse_nodes_in_dist_params(self):
        """
        Returns the nodes names passed as params to the node distribution
        """
        nodes_in_bn = self.network.get_nodes_names()
        nodes_in_params = [n for n in self.distribution_params.split(", ")
                           if n in nodes_in_bn]
        return(nodes_in_params)

    def get_engine_object(self, parents={}, reconstruct=False, save=True):
        # Currently Bayespy only
        if self.engine_object and not reconstruct:
            return self.engine_object
        if self.node_type == self.NODE_TYPE_STOCHASTIC:
            node_distribution = getattr(bp.nodes, self.distribution)
            nodes_in_bn = self.network.get_nodes_names()
            params = []
            kwparams = {'name': self.name}
            # Params currently are only ", "-separated and can be scalars
            # or nodes' names
            for p in self.distribution_params.split(", "):
                if p not in parents:  # nodes_in_bn:
                    params.append(float(p))
                else:
                    # node = BayesianNetworkNode.objects.get(name=p)
                    params.append(parents[p])
            if self.is_observable:
                kwparams['plates'] = (self.ref_model.model_class()
                                      .objects.count(), )
            # import ipdb; ipdb.set_trace()
            self.engine_object = node_distribution(*params, **kwparams)
            # Once initialized, if observable, load the data
            if self.is_observable:
                data = self.get_data()
                self.engine_object.observe(data)
            if save:
                self.engine_object_timestamp = timezone.now()
                self.save()
            return(self.engine_object)

    def get_engine_inferred_object(self, recalculate=False, save=True):
        Q = self.network.perform_inference(recalculate=recalculate)
        self.engine_inferred_object = Q[self.name]
        if save:
            self.engine_inferred_object_timestamp = timezone.now()
            self.save()
        return(self.engine_inferred_object)

    def reset_engine_object(self, save=True):
        self.engine_object = None
        self.engine_object_timestamp = None
        if save:
            self.save()
        return(True)

    def reset_inference(self, save=True):
        self.engine_inferred_object = None
        self.engine_inferred_object_timestamp = None
        self.image.delete()
        if save:
            self.save()
        return(True)

    def update_image(self):
        if (not self.engine_inferred_object or
                not self.node_type == self.NODE_TYPE_STOCHASTIC or
                not self.graph_interval):
            return False
        if self.image:
            self.image.delete()

        a, b = self.graph_interval.split(", ")
        bp.plot.pdf(self.engine_inferred_object,
                    np.linspace(float(a), float(b), num=100), name=self.name)
        image_name = "{0}/{1}".format(
            os.path.join("django_ai",
                         "bayesian_networks"),
            self.network.name + "_" + self.name + ".png")
        bp.plot.pyplot.savefig(settings.MEDIA_ROOT + '/' + image_name)
        bp.plot.pyplot.close()
        self.image = image_name
        self.save()

    def __str__(self):
        return(str(self.network) + " - " + self.name)

    class Meta:
        unique_together = ["network", "name", ]


def update_bn_image(sender, **kwargs):
    kwargs['instance'].network.update_image()

post_save.connect(update_bn_image, sender=BayesianNetworkNode)
post_save.connect(update_bn_image, sender=BayesianNetworkEdge)



# -*- coding: utf-8 -*-

DIST_GAUSSIAN = "Gaussian"  # Gaussian(mu, Lambda, \*\*kwargs)
DIST_GAUSSIAN_ARD = "GaussianARD"  # GaussianARD(mu, alpha[, ndim, shape])
DIST_GAMMA = "Gamma"  # Gamma(a, b, **kwargs)
DIST_WISHART = "Wishart"  # Wishart(n, V, **kwargs)
DIST_EXPONENTIAL = "Exponential"  # Exponential(l, **kwargs)
DIST_GAUSSIAN_GAMMA = "GaussianGamma"  # GaussianGamma(*args, **kwargs)
DIST_GAUSSIAN_WISHART = "GaussianWishart"  # GaussianWishart(*args, **kw)
DIST_BERNOULLI = "Bernoulli"  # Bernoulli(p, **kwargs)
DIST_BINOMIAL = "Binomial"  # Binomial(n, p, **kwargs)
DIST_CATEGORICAL = "Categorical"  # Categorical(p, **kwargs)
DIST_MULTINOMIAL = "Multinomial"  # Multinomial(n, p, **kwargs)
DIST_POISON = "Poisson"  # Poisson(l, **kwargs)
DIST_BETA = "Beta"  # Beta(alpha, **kwargs)
DIST_DIRICHLET = "Dirichlet"  # Dirichlet(*args, **kwargs)
# Other Stochastics Nodes (like point estimators) - Should be Stoch Type?
DIST_MIXTURE = "Mixture"  # Mixture(z, node_class, *params[, cluster_plat])
DIST_MAXIMUMLIKELIHOOD = "MaximumLikelihood"  # MaximumLikelihood(arr[, r])
DIST_CONCENTRATION = "Concentration"  # Concentration(D[, regularization])
DIST_GAMMA_SHAPE = "GammaShape"  # GammaShape(**kwargs)
DISTRIBUTION_CHOICES = (
    (DIST_GAUSSIAN, DIST_GAUSSIAN),
    (DIST_GAUSSIAN_ARD, DIST_GAUSSIAN_ARD),
    (DIST_GAMMA, DIST_GAMMA),
    (DIST_WISHART, DIST_WISHART),
    (DIST_EXPONENTIAL, DIST_EXPONENTIAL),
    (DIST_GAUSSIAN_GAMMA, DIST_GAUSSIAN_GAMMA),
    (DIST_GAUSSIAN_WISHART, DIST_GAUSSIAN_WISHART),
    (DIST_BERNOULLI, DIST_BERNOULLI),
    (DIST_BINOMIAL, DIST_BINOMIAL),
    (DIST_CATEGORICAL, DIST_CATEGORICAL),
    (DIST_MULTINOMIAL, DIST_MULTINOMIAL),
    (DIST_POISON, DIST_POISON),
    (DIST_BETA, DIST_BETA),
    (DIST_DIRICHLET, DIST_DIRICHLET),
    (DIST_MIXTURE, DIST_MIXTURE),
    (DIST_MAXIMUMLIKELIHOOD, DIST_MAXIMUMLIKELIHOOD),
    (DIST_CONCENTRATION, DIST_CONCENTRATION),
    (DIST_GAMMA_SHAPE, DIST_GAMMA_SHAPE),
)
DET_DOT = "Dot"  # Dot(*args, **kwargs)
DET_SUM_MULTIPLY = "SumMultiply"  # SumMultiply(*args[, iterator_axis])
DET_ADD = "Add"  # Add(*nodes, **kwargs)
DET_GATE = "Gate"  # Gate(Z, X[, gated_plate, moments])
DET_TAKE = "Take"  # Take(node, indices[, plate_axis])
DET_FUNCTION = "Function"  # Function(function, *nodes_gradients[, shape])
DET_CONCAT_GAUSSIAN = "ConcatGaussian"  # ConcatGaussian(*nodes, **kwargs)
DET_CHOOSE = "Choose"  # Choose(z, *nodes)
DETERMINISTIC_CHOICES = (
    (DET_DOT, DET_DOT),
    (DET_SUM_MULTIPLY, DET_SUM_MULTIPLY),
    (DET_ADD, DET_ADD),
    (DET_GATE, DET_GATE),
    (DET_TAKE, DET_TAKE),
    (DET_FUNCTION, DET_FUNCTION),
    (DET_CONCAT_GAUSSIAN, DET_CONCAT_GAUSSIAN),
    (DET_CHOOSE, DET_CHOOSE),
)



# -*- coding: utf-8 -*-

from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^run_inference/(?P<bn_id>[0-9]+)$',
        views.bn_run_inference,
        name="bn_run_inference"),
    url(r'^reset_inference/(?P<bn_id>[0-9]+)$',
        views.bn_reset_inference,
        name="bn_reset_inference"),
]



# -*- coding: utf-8 -*-

from django.shortcuts import (redirect, get_object_or_404)
from django.contrib.auth.decorators import user_passes_test

from .models import BayesianNetwork


@user_passes_test(lambda u: u.is_superuser)
def bn_run_inference(request, bn_id):
    bn = get_object_or_404(BayesianNetwork, pk=bn_id)
    bn.perform_inference(recalculate=True)

    return redirect(request.META.get('HTTP_REFERER', '/'))


@user_passes_test(lambda u: u.is_superuser)
def bn_reset_inference(request, bn_id):
    bn = get_object_or_404(BayesianNetwork, pk=bn_id)
    bn.reset_inference()

    return redirect(request.META.get('HTTP_REFERER', '/'))



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-04-26 22:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='UserInfo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('age', models.IntegerField(verbose_name='Age')),
                ('sex', models.CharField(choices=[(0, 'M'), (1, 'F')], max_length=1, verbose_name='Sex')),
            ],
        ),
    ]



# coding: utf-8

from django.contrib import admin

from .models import (BayesianNetwork, BayesianNetworkNode, BayesianNetworkEdge)


class BayesianNetworkNodeInline(admin.StackedInline):
    model = BayesianNetworkNode
    extra = 1


class BayesianNetworkEdgeInline(admin.TabularInline):
    model = BayesianNetworkEdge
    extra = 1


@admin.register(BayesianNetwork)
class BayesianNetworkAdmin(admin.ModelAdmin):
    readonly_fields = ['image', ]
    inlines = [
        BayesianNetworkNodeInline,
        BayesianNetworkEdgeInline,
    ]


# @admin.register(BayesianNetworkNode)
# class BayesianNetworkNodeAdmin(admin.ModelAdmin):
#     pass


# @admin.register(BayesianNetworkEdge)
# class BayesianNetworkNodeEdge(admin.ModelAdmin):
#     pass



# -*- coding: utf-8 -*-

from django.contrib import admin

from .models import UserInfo


@admin.register(UserInfo)
class UserInfoAdmin(admin.ModelAdmin):
    list_display = ['id', 'age', 'get_sex_display', 'avg1']



# -*- coding: utf-8 -*-

from django.db import models


class UserInfo(models.Model):
    """
    Example User Information Model
    """
    SEX_CHOICES = ((0, "M"), (1, "F"))

    age = models.IntegerField("Age")
    sex = models.CharField("Sex", choices=SEX_CHOICES, max_length=1)
    avg1 = models.FloatField("Average 1", blank=True, null=True)

    def __unicode__(self):
        return(self)
